#ifndef BTREE_DEFINITIONS_INCLUDED
#define BTREE_DEFINITIONS_INCLUDED

#ifdef __cplusplus
const char BTREE_rcsid[] = "$Id: BTree.h,v 1.2 1998/04/23 18:10:41 mmorin Exp $";
#else
static const char BTREE_rcsid[] = "$Id: BTree.h,v 1.2 1998/04/23 18:10:41 mmorin Exp $";
#endif

/******************************************************************************
 *
 * (c) Copyright 1992-1998 Scheduling Systems Incorporated
 * All rights reserved.
 *
 * These programs are supplied under a license.  They may not be used,
 * disclosed, and/or copied except as permitted under such license
 * agreement.  Any copy must contain the above copyright notice and
 * this notice.  Use, copying, and/or disclosure of the programs is
 * strictly prohibited unless otherwise provided in the license agreement.
 *
 ******************************************************************************
 *
 * $Author: mmorin $  Created By: mmorin
 * $Date: 1998/04/23 18:10:41 $  Creation: Fri Apr  3 11:31:56 1998
 * $Revision: 1.2 $
 * $State: Exp $
 *
 * Revisions by:
 *	Ahmet M. Buharali		(amb, ahmet)
 *	William P. Hees			(wph)
 *      Saeed Honaryar                  (sh)
 *	Jeffrey Donner			(jgd, JeffD)
 *     	John D. Pararas                 (jdp)
 *      David J. Picard                 (djp, dave)
 *
 * $Log: BTree.h,v $
 * Revision 1.2  1998/04/23  18:10:41  mmorin
 * Moved implementation of the template into the .h
 *
 * Revision 1.1  1998/04/21  20:51:15  mmorin
 * Initial revision
 *
 *
 *****************************************************************************/

//
// BTree HOWTO!
//
// by mmorin
//
//
// This is a template that implement the BTree data structure.
// This file depend on BTreeNode.{cpp,h} that implement the node
// itself.
// For compiling the test for BTreeNode use 
//  g++  -g  -D__DEBUG_BTREENODE__ BTreeNode.cpp -o btreenode  
// For compiling and testing BTree use
//  g++  -g  -D__DEBUG_BTREE__ BTree.cpp -o btree  
// The way of using BTree is to include the BTree.h into your source
// file. 
//
// For creating a BTree of any type declare
//  BTree<TYPE> btree;
// For the data of type TYPE you have to define operators =, ==, <
// (the last one is not strictly necessary).
// You can specify the number 2*t+1 of element in a node of a BTree: this
// fix the number of step in a search (more higher is this number less
// "search step" are required, but more wasted space could be present).
// For fixind that you figure out the number of element you want to
// put in the btree: with t=10 you can index into 2 step 
// 21 + (21+1) * 21 + (21+1) * (21+1) * 21 elements = 10647 elements
// so you could fix the number of steps (eg 3) and the number of
// elmenent (eg 10000) and find out t =~ log3 5000 =~ 8
// For default t is fixed to 10
// Be careful when you use pointers of a class: in that case you have
// to build a ptrClass wraparound: it has to overload the operator = <
// == with calling the right method of the data you are using.
//


#include "BTreeNode.h"

template<class TYPE>
class BTree {
private:
  BTreeNode<TYPE> *m_head;	// the head of the BTree

  int m_height;			// the levels
  int m_nKeys;			// the number of key present

  BTreeNode<TYPE> *m_currNode;	// the node of the previous search
  
  // Private methods
  void FreeBTree (BTreeNode<TYPE>*);

  bool SearchInBTree(BTreeNode<TYPE>*, TYPE &element);

  void SplitChild (BTreeNode<TYPE>*, int, BTreeNode<TYPE>*);
  void InsertInBTreeNonFullNode (const TYPE &, BTreeNode<TYPE>*&);

  bool DeleteFromBTree(const TYPE &, BTreeNode<TYPE>*);
  void DeleteFromBTreeInternalNode(BTreeNode<TYPE>*, int);
  void MergeChild(BTreeNode<TYPE>* &, TYPE &, BTreeNode<TYPE>*);

  void VisitBTree (ostream&, const BTreeNode<TYPE>*, int indent = -1) const;
public:
  
  BTree(int arity=1);
  BTree(const BTree<TYPE>& btree);
  ~BTree();

  int Arity () { return m_head->Arity() / 2; }

  bool Search (TYPE& element) { return SearchInBTree(m_head, element); }
  void Insert (const TYPE& element);
  bool Delete (const TYPE& element);

  // walking on the tree
  bool Next(TYPE&);
  bool Prev(TYPE&);

  TYPE First();
  TYPE Last();    
  
  friend ostream& operator<< (ostream &, const BTree<TYPE>& );
  void Save (ofstream &os) {
      VisitBTree(os, m_head);
  }
  int GetHeight() { return m_height; }
};

// -----------------------
// TEMPLATE IMPLEMENTATION
// -----------------------


/*
 *------------------------------------------------------------------------
 *
 *  BTree::BTree --
 *
 *  Create the BTree: note that t is the minimum number of element 
 *  on each node. 
 *  We will refer for complexity calculation:
 *  <t> for the arity 
 *  <n> the total number of element
 *  <h> for the height of the tree (by def Logt(n) where Logt is logarithm base t) 
 *  Complexity:
 *    O(1) constant
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTree<TYPE>::BTree(int t) {

  m_head = new BTreeNode<TYPE>(2 * t + 1);
  m_nKeys = 0;
  m_height = 1;
  m_currNode = m_head;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::BTree --
 *
 *  Copy constructor: note the tree is not full copied; only the head is
 *  moved (for our purpose it could be enought).
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTree<TYPE>::BTree (const BTree<TYPE>& btree) {
  m_head     = btree.m_head;
  m_nKeys    = btree.m_nKeys;
  m_height   = btree.m_height;
  m_currNode = btree.m_currNode;
}


/*
 *------------------------------------------------------------------------
 *
 *  BTree::~BTree --
 *
 *  Destructor: recursively destroy the tree.
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTree<TYPE>::~BTree () {
  FreeBTree(m_head);
  m_head = NULL;		// it could be not necessary
}



/*
 *------------------------------------------------------------------------
 *
 *  BTree::FreeBTree -- PRIVATE 
 *
 *  Destroy each node of the BTree in recursive way.
 *  Not that we need only to free ptr, because all the sons will be freed
 *  in the deep recursion. This will free the head too!
 *  Note the +1 for accessing the last son of the node!
 *  Complexity 0(n)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTree<TYPE>::FreeBTree (BTreeNode<TYPE>* ptr ) {
  if (ptr != NULL) {
    if (!ptr->IsLeaf()) {
      for (int i=0; i < ptr->Size() + 1; i++) {
	FreeBTree(ptr->Left(i));
      }
    }
    delete ptr;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::SearchInBTree -- PRIVATE
 *
 *  Start a search in the tree from the ptr.
 *  Parameter:
 *   ptr      (input) is the node where to search the data 
 *   element  (i/o) contais the key to search; if found will contain the 
 *            data
 *  Returned value:
 *   true if succefully found, false otherwise. 
 *  Complexity: it is called at maximum O(h)
 *              the while is O(t)
 *              in total it is O(th) = O(t Logt(n))
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
bool
BTree<TYPE>::SearchInBTree(BTreeNode<TYPE> *ptr, 
		     TYPE &element) {
  bool found = false;

  int i = ptr->FindPosition(element);

  if (i < ptr->Size() && element == ptr->GetNth(i)) {
    element = ptr->GetNth(i);
    // Ok start updating position 
    // (the rest is recursively done at the end)
    ptr->SetPos(i);    
    m_currNode = ptr;
    return true;
  }
  
  if (ptr->IsLeaf()) {
    return false;
  } else {

    found = SearchInBTree(ptr->Left(i), element);        

    if (found) {
      // update all position on cascade for linear search
      ptr->SetPos(i-1);		
    }
    return found;
  }
}

  
/*
 *------------------------------------------------------------------------
 *
 *  BTree::Insert -- 
 *
 *  Insert an element into the BTree; It will start the recursion!
 *  This part will handle the root of the tree to be full;
 * 
 *  Parameter:
 *   element  (input by ref) is the element to insert
 *   ptr      (input/output) the node where to insert
 *
 *  Returned value:
 *   none
 *  Side effect:
 *   The insertion of the element on the head, and start recursion 
 *   (increment the height).
 *  Complexity: 
 *   O(th) 
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTree<TYPE>::Insert (const TYPE &element) {

  if (m_head->IsFull()) {
    m_height++;

    BTreeNode<TYPE> *newOne = new BTreeNode<TYPE>(2 * Arity() + 1);
    newOne->SetLeaf(false);
    newOne->Left(0) = m_head;
    
    SplitChild(newOne, 0, m_head);    
    
    m_head = newOne;
  }

  InsertInBTreeNonFullNode(element, m_head);

  m_nKeys++;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::SplitChild -- PRIVATE
 *
 *  Divide a full node into two node (half full)
 *  Parameter:
 *   father   (input) is the node father of the node we want to split
 *   position (input) the position where to insert the middle element of fullOne
 *   fullOne  (input) the node to be splitted
 *  Returned value:
 *   none
 *  Side effect:
 *   the tree made by father->fullOne will be splitted into 
 *    father -+          +- father -+
 *         fullOne    fullOne     newOne
 *  Complexity: 
 *   O(t) due to the loops
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTree<TYPE>::SplitChild (BTreeNode<TYPE>* father, int position, BTreeNode<TYPE>* fullOne ) {
  
  int j;
  int t = Arity();

  BTreeNode<TYPE>* newOne = new BTreeNode<TYPE>(2 * t + 1);

  // we split on the same level!  
  newOne->SetLeaf( fullOne->IsLeaf() );
  // I put half of the element
  newOne->SetSize(t);
  // I know who is the father for sequential search
  // newOne and fullOne are brother now
  newOne->SetFather(father);
  fullOne->SetFather(father);

  // copy half of the sons
  // newOne will be on th right of fullOne
  for (j = 0; j < t; j++) {
    newOne->GetNth(j) = fullOne->GetNth(t+j+1);
  }

  // copy half of the the links to the sons (plus the right of the last!
  if (!fullOne->IsLeaf()) {
    for (j = 0; j < t+1; j++) {
      newOne->Left(j) = fullOne->Left(t+j+1);
    }    
  }
  fullOne->SetSize(t);

  // start moving element on father
  // I've to move the links from position to the end of the father 
  // of one position on the right
  for (j=father->Size(); j >= position; j--) {
    father->Left(j+1) = father->Left(j);
  }

  // insert the new node
  father->Right(position) = newOne;

  for (j=father->Size(); j > position; j--) {
    father->GetNth(j) = father->GetNth(j-1);
  }

  // insert the new one
  father->GetNth(position) = fullOne->GetNth(t);
  father->SetSize( father->Size() + 1 );
  father->SetPos(position);

  // correct the position link:
  int pos = fullOne->GetPos();
  if (pos >= t) {
    newOne->SetPos(pos-t);
    fullOne->SetPos(0);
  }
  // the father of sons of newOne is newOne, and not his brother 
  if (!newOne->IsLeaf()) {
    for (j=0; j<newOne->Size()+1; j++) {
      newOne->Left(j)->SetFather(newOne);
    }
  }

}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::InsertInBTreeNonFullNode -- PRIVATE
 *
 *  Insert an element into a node in the case the node is not full; 
 *  if it is full, it will spit it and the recall it.
 *
 *  Parameter:
 *   element  (input by ref) is the element to insert
 *   ptr      (input) is the node we want to change
 *  Returned value:
 *   none
 *  Side effect:
 *   The insertion of the element in the node if not full; the modification 
 *   of the links in any case for syncing the element
 *  Complexity: 
 *   O(th) 
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTree<TYPE>::InsertInBTreeNonFullNode (const TYPE &element, BTreeNode<TYPE>* &ptr) {

  int i;
  if (ptr->IsLeaf() ) {
    
    for (i = ptr->Size() - 1; 
	 i >= 0 && element < ptr->GetNth(i);
	 i--) {
      ptr->GetNth(i+1) = ptr->GetNth(i);
    }

    i = i + 1;	

    // I insert -> let's record the position
    ptr->SetPos(i);    
    ptr->GetNth(i) = element;    
    ptr->SetSize( ptr->Size() + 1 );
    m_currNode = ptr;

  } else {
    // I usee a function of the node for finding the next greter
    i = ptr->FindPosition(element);

    // adjust all the position for linear scan
    ptr->SetPos(i-1);

    BTreeNode<TYPE> *son = ptr->Left(i);
    if (son->IsFull()) {
      SplitChild( ptr, i, son );
      if (element > ptr->GetNth(i)) {
	son = ptr->Right(i);
      }
    }
    
    InsertInBTreeNonFullNode(element, son);
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::Delete -- 
 *
 *  Delete an element from the BTree; thi function will balance
 *  the tree and, consequentially, decrease the depth of the tree
 *
 *  Parameter:
 *   element  (input by ref) is the element to delete
 *  Returned value:
 *   tru if the element is deleted (succesfull found e delete!)
 *  Side effect:
 *   It could change the head, and it change the number of element 
 *   in the tree and the height
 *  Complexity: 
 *   this will call only 1 times; the following complaxity discussion 
 *   is for the full delete:
 *   O(th)=O(t logt(n)) 
 *   See the remove will mostly involve the leaf, so on average
 *   te complaxity is very close to O(1)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
bool
BTree<TYPE>::Delete (const TYPE &element) {
  bool found;
  
  found = DeleteFromBTree(element, m_head);
  if (m_head->Size() == 0 && !m_head->IsLeaf()) {    
    BTreeNode<TYPE>* node = m_head->Left(0);
    delete m_head;
    m_head = node;
    m_height--;
  }

  if (found) {
    m_nKeys--;
  }
  return found;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::DeleteFromBTree -- PRIVATE
 *
 *  Delete an element from the BTree; 
 *  The action are: 
 *  1) find the position
 *  2) if present and the node ND is a leaf -> compact it then done
 *  3) if present but the node ND is not a leaf -> remove it from internal 
 *     node (see below, mutual recursion)
 *  4) if not in ND: we found the subtree where the element has to be 
 *     (left of right son of ND -> son of ND (SoND)).
 *     We have to keep the head of SoND full at least of t element, so:
 *     if it is, go on; if not, we merge the head with a brother (brother 
 *     of SoND (BoSoND)) if possible (SoND.Size < 2t+1), otherwise we rotate 
 *     an element from BoSoND into the SoND)
 *     (substitute an element from BoSoND into ND, and move the substitued 
 *     element from ND (the father) into SoND (the current subtree)).
 *     In this case the current node (SoND) will be full enought for 
 *     deleting the requested  element.
 *  5) If not found go into the "right" subtree.

 *  Parameter:
 *   element  (input by ref) is the element to delete
 *   ptr      (input) the head of the subtree where the element has to be
 *            present
 *  Returned value:
 *   true if the element is deleted (succesfull found e delete!), false 
 *   otherwise
 *  Side effect:
 *   This is called recursivelly (and mutually recursivelly) so:
 *   It compacts a node, it could rotate the subtree containing the element
 *  Complexity: 
 *   O(th)=O(t logt(n)) 
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
bool
BTree<TYPE>::DeleteFromBTree(const TYPE &element,
			     BTreeNode<TYPE>* ptr) {

  int i = ptr->FindPosition(element);

  if (i < ptr->Size() && element == ptr->GetNth(i)) {
    if (!ptr->IsLeaf()) {
      DeleteFromBTreeInternalNode(ptr, i);
      return true;
    } else {
      ptr->Compact(i);
      return true;
    }
  } 
  
  if (ptr->IsLeaf()) {
    return false;
  } else {
    BTreeNode<TYPE>* node = ptr->Left(i);

    if (node->Size() <= Arity()) {
      bool merged = false;

      // first of all I try to merge with a brother
      if (i != ptr->Size() && 
	  ptr->Right(i)->Size() <= Arity()) {	

	MergeChild (ptr->Left(i), ptr->GetNth(i), ptr->Right(i));
	// remove the pending element
	ptr->Compact(i);
	merged = true;

      } else if (i != 0 && 
		 ptr->Left(i-1)->Size() <= Arity()) {

	i = i - 1; 
 	MergeChild (ptr->Left(i), ptr->GetNth(i), node);
	node = ptr->Left(i);	// this is used later!

	ptr->Compact(i);
	merged = true;
      }

      // I have no brother that could merge with node
      // I can move an element from a brother into node
      if (!merged) {
	int n;
	BTreeNode<TYPE>* brother;
	// there is a brother that has enough element
	// it could not be 0 = i = Size() (I've to have 0 elmenet!)
	// it could be both brther have Size() > Arity, so get one
	if ( i != 0 ) {
	  i = i - 1;
	  brother = ptr->Left(i);
	  n = brother->Size() - 1;

	  node->ShiftRight();
	  node->GetNth(0) = ptr->GetNth(i);
	  if (!node->IsLeaf()) {
	    node->Left(0) = brother->Right(n);
	  }
	  ptr->GetNth(i) = brother->GetNth(n);
	  
	  brother->SetSize(n);
	} else {
	  brother = ptr->Right(i);
	  n = node->Size();

	  node->GetNth(n) = ptr->GetNth(i);
	  if (!node->IsLeaf()) {
	    node->Right(n) = brother->Left(0);
	  }
	  ptr->GetNth(i) = brother->GetNth(0);	  
	  node->SetSize(n+1);	
	  brother->ShiftLeft();
	}
      }
    }
    DeleteFromBTree(element, node);   
  }

}    

/*
 *------------------------------------------------------------------------
 *
 *  BTree::DeleteFromBTreeInternalNode -- PRIVATE
 *
 *  Delete an element from and internal node ND; 
 *  This opration is necessary for not only deleting the element, but
 *  for keeping the node full at least of t elements.
 *  For doing that we have to see if I can rotate (a brother has enought 
 *  element).
 *  In that case if the brohter is the left one I have to switch the most
 *  right element in the node ND, and the erase it from the subtree (mutual
 *  recursion with DeleteFromBTree, this because I could rebalance the
 *  tree again!); In the case is the right brother one I have to switch
 *  the most left element in the node ND, and the erase it from the
 *  subtree (....);
 *  In the case I can't rotate -> I merge the two sons, then erase the
 *  element from the head and go ahead erasing the element into the
 *  subtree with the new "merged head".
 *
 *  Parameter:
 *   ptr   (input) the node where the element do delete is present
 *   pos   (input) the position in the node ptr of the element to erase
 *  Returned value:
 *   None
 *  Side effect:
 *   This call the DeleteFromBtree (mutual recursion!) The tree
 *   structure is changed: it will rotate or shrinked.
 *
 *------------------------------------------------------------------------
 */
template<class TYPE> 
void
BTree<TYPE>::DeleteFromBTreeInternalNode(BTreeNode<TYPE>* ptr,
					 int pos) {

  BTreeNode<TYPE>* node;
  int t = Arity();
  TYPE element = ptr->GetNth(pos);

  node = ptr->Left(pos);
  if (node->Size() > t) {
    // Ok I ken compact    
    node->SetPos( node->Size() - 1 );
    BTreeNode<TYPE>* right = node->AllRight();
    TYPE biggest = right->GetNth(right->Size() - 1);
    DeleteFromBTree(biggest, node);
    ptr->GetNth(pos) = biggest;
    return;
  }

  node = ptr->Right(pos);
  if (node->Size() > t) {
    // Ok I can compact
    node->SetPos(0);
    BTreeNode<TYPE>* left = node->AllLeft();
    TYPE smallest = left->GetNth(0);
    DeleteFromBTree(smallest, node);
    ptr->GetNth(pos) = smallest;
    return;
  }

  MergeChild(ptr->Left(pos), 
	     ptr->GetNth(pos),
	     ptr->Right(pos));

  ptr->Compact(pos);
  DeleteFromBTree(element, ptr->Left(pos));
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::MergedChild -- PRIVATE
 *
 *  Merge to child on the left.
 *
 *  Parameter:
 *   merged  (input/output) the left node, and result of the merge
 *   element (input) the middle element
 *   node    (input) the right node to be merged
 *  Returned value:
 *   None
 *  Side effect:
 *   None
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTree<TYPE>::MergeChild(BTreeNode<TYPE>* &merged, 
		       TYPE &element,
		       BTreeNode<TYPE>* node) {

  int n = merged->Size();
  
  // copy the key to be erased
  merged->GetNth(n) = element;
  
  // merging
  for (int i=0; i < node->Size(); i++ ) {
    merged->GetNth(n + i + 1) = node->GetNth(i);
  }
  merged->SetSize(n + node->Size() + 1);

  // merge the links if any: remember merged and node are brothers
  if (!merged->IsLeaf()) {
    for (int i=0; i < node->Size(); i++ ) {
      merged->Left(n + i + 1) = node->Left(i);
    }
    merged->Left(merged->Size()) = node->Left(node->Size());
  }
  
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::First -- 
 *
 *  Return the First (smallest) element in the BTree.
 *
 *  Parameter:
 *   none
 *  Returned value:
 *   The smallest element
 *  Side effect:
 *   Change all the position in the left branch of the tree to point to
 *   the first element!
 *   Now the position (current node) is the left deepest
 *  Complexity: 
 *   O(h)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
TYPE
BTree<TYPE>::First() {
  m_currNode = m_head;
  while (!m_currNode->IsLeaf()) {
    m_currNode->SetPos(0);
    m_currNode = m_currNode->Left(0);
  }
  m_currNode->SetPos(0);
  return m_currNode->GetNth(0); 
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::Last -- 
 *
 *  Return the Last (largest) element in the BTree.
 *
 *  Parameter:
 *   none
 *  Returned value:
 *   The largest  element
 *  Side effect:
 *   Change all the position in the right branch of the tree to point to
 *   the last element!
 *   Now the position (current node) is the right deepest
 *  Complexity: 
 *   O(h)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
TYPE
BTree<TYPE>::Last() {
  m_currNode = m_head;
  int n;
  while (!m_currNode->IsLeaf()) {
    n = m_currNode->Size() - 1;
    m_currNode->SetPos(n);
    m_currNode = m_currNode->Right(n); 
  }
  n = m_currNode->Size() - 1;
  m_currNode->SetPos(n);
  return m_currNode->GetNth(n); 
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::Next -- 
 *
 *  Return the next element after the previous search or insert.
 *  I need a variable that say me if in a node I've to go down in 
 *  the subtree or left on the nex availbale element.
 *  This variable is m_prevPos: if previusly I was on the last element
 *  of a node, when I rise up, I have to print the root, otherwise I've
 *  to go down on the right subtree!
 *
 *  Parameter:
 *   element  (output) is the element after the last searched
 *  Returned value:
 *   true if exist a next element, false otherwise
 *  Side effect:
 *   It change the current node where actiong on the search
 *  Complexity: 
 *   O(1): note recursion can be called at maximum O(h)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
bool
BTree<TYPE>::Next(TYPE &element) {
  BTreeNode<TYPE>* node = m_currNode;  
  int pos = m_currNode->GetPos();

  if (node->IsLeaf()) {
    if (pos != node->Size() - 1) {
      pos ++;
      element = node->GetNth(pos);
      node->SetPos(pos);
      return true;
    } else {
      node = node->UpRight();
      if (node == NULL) {
	return false;
      }
      pos = node->GetPos();
      element = node->GetNth(pos);
      m_currNode = node;
      return true;
    }
  } else {
    node = node->Right(pos);
    node->SetPos(0);
    m_currNode = node->AllLeft();
    element = m_currNode->GetNth(0);
    return true;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::Prev -- 
 *
 *  Return the prev element before the previous found in a search or insert.
 *
 *  Parameter:
 *   element  (output) is the element before the last searched
 *  Returned value:
 *   true if a precedent element exists, false otherwise
 *  Side effect:
 *   It change the current node where actiong on the search
 *  Complexity: 
 *   O(1): note recursion can be called at maximum O(h)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
bool
BTree<TYPE>::Prev(TYPE &element) {
  BTreeNode<TYPE>* node = m_currNode;  
  int pos = m_currNode->GetPos();

  if (node->IsLeaf()) {
    if (pos != 0) {
      pos --;
      element = node->GetNth(pos);
      node->SetPos(pos);
      return true;
    } else {
      node = node->UpLeft();
      if (node == NULL) {
	return false;
      }
      pos = node->GetPos();
      element = node->GetNth(pos);
      m_currNode = node;
      return true;
    }
  } else {
    node = node->Left(pos);
    node->SetPos(node->Size() - 1);
    m_currNode = node->AllRight();
    element = m_currNode->GetNth(m_currNode->Size() - 1);
    return true;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::operator<< --
 *
 *  Display the tree in graphic way.
 *
 *  Parameter:
 *   os    (input by ref) is the channel where to display
 *   bt    (input) the tree to display
 *  Returned value:
 *   the channel
 *  Side effect:
 *   none
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
ostream& 
operator<< (ostream &os, const BTree<TYPE>& bt) {

  os << "[" << bt.m_height << "," << bt.m_nKeys << "] =\n";
  
  bt.VisitBTree(os, bt.m_head, 0);
  return os;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTree::VisitBTree -- PRIVATE
 *
 *  Visit in preorder way the tree, and display it in a graph way
 *
 *  Parameter:
 *   os     (input by ref) is the channel where to display
 *   node   (input) the node to display
 *   indent (input) is the amount of space needed for indenting the tree
 *          for a "pretty" looking       
 *  Returned value:
 *   none
 *  Side effect:
 *   none
 *  Complexity:
 *   O(n) I scan all the tree
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
void
BTree<TYPE>::VisitBTree (ostream &os, const BTreeNode<TYPE> *node, int indent) 
const 
{
  int i;
  if (indent != -1 ) {
    for (i=0; i < indent; i++) {
      os << "| "; 
    }
  }

  node->Save(os);
  if (!node->IsLeaf()) {
    for (i=node->Size(); i >= 0; i--) {
      VisitBTree(os, node->Left(i), (indent == -1 ? -1 : indent+i));
    }
  }
}

#endif /* BTREE_DEFINITIONS_INCLUDED */
