#ifndef BTREENODE_DEFINITIONS_INCLUDED
#define BTREENODE_DEFINITIONS_INCLUDED

#ifdef __cplusplus
const char BTREENODE_rcsid[] = "$Id: BTreeNode.h,v 1.2 1998/04/23 18:12:29 mmorin Exp $";
#else
static const char BTREENODE_rcsid[] = "$Id: BTreeNode.h,v 1.2 1998/04/23 18:12:29 mmorin Exp $";
#endif

/******************************************************************************
 *
 * (c) Copyright 1992-1998 Scheduling Systems Incorporated
 * All rights reserved.
 *
 * These programs are supplied under a license.  They may not be used,
 * disclosed, and/or copied except as permitted under such license
 * agreement.  Any copy must contain the above copyright notice and
 * this notice.  Use, copying, and/or disclosure of the programs is
 * strictly prohibited unless otherwise provided in the license agreement.
 *
 ******************************************************************************
 *
 * $Author: mmorin $  Created By: mmorin
 * $Date: 1998/04/23 18:12:29 $  Creation: Mon Apr  6 10:32:58 1998
 * $Revision: 1.2 $
 * $State: Exp $
 *
 * Revisions by:
 *	Ahmet M. Buharali		(amb, ahmet)
 *	William P. Hees			(wph)
 *      Saeed Honaryar                  (sh)
 *	Jeffrey Donner			(jgd, JeffD)
 *     	John D. Pararas                 (jdp)
 *      David J. Picard                 (djp, dave)
 *
 * $Log: BTreeNode.h,v $
 * Revision 1.2  1998/04/23  18:12:29  mmorin
 * Moved implementation into .h
 * changed Move to use a non const pos
 * Add all the includes on the head
 *
 * Revision 1.1  1998/04/21  20:51:22  mmorin
 * Initial revision
 *
 *
 *****************************************************************************/

#include <iostream.h>
#include <string.h>
#include <stdlib.h>


#define ASSERT(ptr) {\
  if (ptr == NULL) {\
      cerr << "Can't allocate data for BTreeNode\n";\
      abort();\
    }\
  }

template<class TYPE> 
class BTreeNode {
private:
  TYPE* m_keys;			// the real element
  BTreeNode<TYPE> **m_links;	// all the pointer to the sons: if NULL it is a leaf
  int  m_nElement;		// the element present in the node
  int  m_arity;			// the minimum number of number of element in the node
				//  ( it is the <t> value )
  BTreeNode<TYPE>* m_father;	// the father node
  int m_currPos;		// the current position on node
  
  void SetInternalNode();
  void Copy(const BTreeNode<TYPE>& node);
  void Move(int n, const int versus, const bool compactRight = false);
  //  static int Compare (const TYPE* a, const TYPE* b);

public:
  BTreeNode(int arity = 31);
  BTreeNode(const BTreeNode<TYPE>& node);
  ~BTreeNode();

  // accessing members
  bool IsFull() const              { return (m_nElement == m_arity); }

  int GetPos() const               { return m_currPos;}
  void SetPos(int n)               { m_currPos = n; }
  
  BTreeNode<TYPE>* Father() const    { return m_father; }
  void SetFather(BTreeNode<TYPE>* p) { m_father = p; }

  bool IsLeaf() const              { return m_links == NULL; }
  void SetLeaf(bool v = true)      { if (!v) SetInternalNode(); }

  int Size() const                 { return m_nElement; }
  void SetSize(int n)              { m_nElement = n; }

  TYPE& operator[](int n)          { return m_keys[n]; }
  TYPE& GetNth(int n)              { return m_keys[n]; }

  int Arity() const                { return m_arity;  }
  void SetArity(int arity)         { m_arity = arity; }

  // accessing sons: REMEMBER it has not to be a leaf!
  BTreeNode<TYPE>* & Left(int n) const  { return m_links[n];  }
  BTreeNode<TYPE>* & Right(int n) const { return m_links[n+1]; }

  BTreeNode<TYPE>* AllLeft();
  BTreeNode<TYPE>* AllRight();
  BTreeNode<TYPE>* UpRight();
  BTreeNode<TYPE>* UpLeft();

  // operator
  BTreeNode<TYPE>& operator=(const BTreeNode<TYPE>& node);
  friend ostream& operator<<(ostream &os, const BTreeNode<TYPE> &node);

  int FindPosition(const TYPE &element); 
  void Append(const TYPE &element);
  void Display(ostream& os=cout);
  
  void ShiftRight()         { Move(0, +1); }
  void ShiftLeft()          { Move(0, -1); }
  void Compact(const int n) { 
    if (m_links != NULL) {
      delete m_links[n+1];
    }
    Move(n, -1, true); 
  }
  void Save (ostream &os);
};

// -----------------------
// TEMPLATE IMPLEMENTATION
// -----------------------

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::BTreeNode --
 *
 *  Allocate the node: see that the links have 1 element more than the keys
 *  Note that the 2 array have reserved some space that not necessarelly is 
 *  identical ad arity!!!
 *  Note we are optimizing the space: in the beginning a node is a leaf
 *  so it is not necessary to allocate space for the sons!
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
BTreeNode<TYPE>::BTreeNode(int arity) {
  m_nElement = 0;
  m_arity    = arity;
  m_father   = NULL;
  m_currPos  = 0;

  m_links    = NULL;		// this means it is a leaf!

  // allocate and check
  m_keys  = new TYPE[m_arity];
  ASSERT(m_keys);
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::BTreeNode --
 *
 *  Allocate the node: this is a copy constructor THIS IS NEEDED BY STL
 *  Note we are optimizing the space: in the beginning a node is a leaf
 *  so it is not necessary to allocate space for the sons! We will see
 *  in the copy if <node> has sons or not
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
BTreeNode<TYPE>::BTreeNode(const BTreeNode<TYPE>& node) {  
  // initialize all the pointers!
  m_keys = NULL;
  m_links = NULL;

  Copy(node);
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::Copy --
 *
 *  Copy a node into the current one.
 *  Note we are optimizing the space: if a node is a leaf it is not necessary to 
 *  allocate space for the sons! (it is setted a NULL in the constructor)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
void
BTreeNode<TYPE>::Copy(const BTreeNode<TYPE>& node) {
  int i;

  m_nElement = node.m_nElement;
  m_arity    = node.m_arity;
  m_father   = node.m_father;
  m_currPos  = node.m_currPos;

  // allocate the keys
  if (m_keys == NULL) {
    m_keys  = new TYPE[m_arity];
    ASSERT(m_keys);
  }

  //  copy all the keys
  for(i=0; i < m_nElement; i++) {
    m_keys[i] = node.m_keys[i];
  }

  // copy, eventually, all the links
  if (node.m_links != NULL) {
    SetInternalNode();
    for(i=0; i < m_nElement+1; i++) {
      m_links[i] = node.m_links[i];
    }
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::~BTreeNode --
 *
 *  Destructors: deallocate space
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
BTreeNode<TYPE>::~BTreeNode() {
  delete[] m_keys;
  if (m_links != NULL) {
    delete[] m_links;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::SetInternalNode --
 *
 *  Create the links to other node if not already present.
 *  Note we are optimizing the space: this time we have to
 *  allocate space for the sons if not alredy present!
 *
 *------------------------------------------------------------------------
 */
template<class TYPE> 
void 
BTreeNode<TYPE>::SetInternalNode() {
  if (m_links == NULL) {
    m_links = new BTreeNode<TYPE>*[m_arity + 1];
    ASSERT(m_links);
    for(int i=0; i < m_arity + 1; i++) {
      m_links[i] = NULL;
    }
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::operator= --
 *
 *  Copy the node into another
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
BTreeNode<TYPE>&
BTreeNode<TYPE>::operator=(const BTreeNode<TYPE>& node) {

  if (&node == this) {
    return *this;
  }

  // free links if <node> is a leaf
  if (node.m_links == NULL && m_links != NULL) {
    delete[] m_links;
    m_links = NULL;
  }

  Copy(node);
  return *this;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::UpLeft --
 *
 *  This function is releated to the BTree.
 *  Rise the tree till the current node is a left son
 *  Be careful: the search in the BTree is done in the way it recursively
 *  reset the position to be the precedent one. Note so tha if we are 
 *  on the first position m_currPos will be -1, and that is handled 
 *  by the first if!
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTreeNode<TYPE>*
BTreeNode<TYPE>::UpLeft() {  
  BTreeNode<TYPE> *father = m_father;
  
  if (father == NULL) {
    return NULL;
  }
  
  int pos = father->m_currPos;

  if (pos == -1) {
      return father->UpLeft();
  } else if (father->Left(pos) == this) {
    if (pos == 0) {
      return father->UpLeft();
    } else {
      father->SetPos(pos-1);
      return father;
    }
  } else {    
    return father;
  }
}


/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::Move -- PRIVATE
 *
 *   This will move the contents of a node a position right (if
 *   versus = +1) or a position left (if versus = -1) starting
 *   from position pos.
 *   Note the son created will be NULL; the son shrinked is NOT deleted
 *   This will be freed on if it is a compact action and not a shift 
 *   action!
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
void
BTreeNode<TYPE>::Move(int pos, 
		      const int versus, 
		      const bool compactRight) {  
  int i;
  
  // move the contents
  if (versus == -1 ){
    for (i=pos; i < m_nElement-1; i++)  m_keys[i] = m_keys[i+1];
  } else {
    for (i=m_nElement; i>pos; i--)  m_keys[i] = m_keys[i-1];
  }

  // move the links if any
  if (m_links != NULL) {    
    if (versus == -1 ){
      // I'm compacting: erase the son and shrink

      // I want to erase the right son, no the left!
      if (compactRight) {
	pos = pos + 1;
      }
      for (i=pos; i < m_nElement; i++)  m_links[i] = m_links[i+1];
    } else {
      // I'm adding one: expand and reset the son
      for (i=m_nElement+1; i>pos; i--)  m_links[i] = m_links[i-1];
      m_links[pos] = NULL;
    }
  }

  m_nElement = m_nElement + versus;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::UpRight --
 *
 *  This function is releated to the BTree.
 *  Rise the tree till the current node is a right son
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTreeNode<TYPE>*
BTreeNode<TYPE>::UpRight() {  
  BTreeNode<TYPE> *father = m_father;
  
  if (father == NULL) {
    return NULL;
  }
  
  int pos = father->m_currPos;

  if (father->Right(pos) == this) {
    if (pos == father->Size() - 1) {
      return father->UpRight();
    } else {
      father->SetPos(pos+1);
      return father;
    }
  } else {    
    return father;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::AllLeft --
 *
 *  This function is releated to the BTree.
 *  Go Dowun in the tree along the left son
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTreeNode<TYPE>*
BTreeNode<TYPE>::AllLeft() {  

  if (m_links != NULL) {
    BTreeNode<TYPE> *next = Left(m_currPos);
    next->SetPos(0);
    return next->AllLeft();
  } else {
    return this;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::AllRight --
 *
 *  This function is releated to the BTree.
 *  Go Dowun in the tree along the right son
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>
BTreeNode<TYPE>*
BTreeNode<TYPE>::AllRight() {  

  if (m_links != NULL) {
    BTreeNode<TYPE> *next = Right(m_currPos);
    next->SetPos(next->Size()-1);
    return next->AllRight();
  } else {
    return this;
  }
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::FindPosition --
 *
 *  Find the position of an element, if present, or the less greater one
 *  (e can use a binary search!)
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
int
BTreeNode<TYPE>::FindPosition(const TYPE& element) {
  int i = 0;
//   TYPE* pos;
//   pos = bsearch(&element, m_keys, m_nElement, sizeof(TYPE),  
  while ( i < m_nElement && m_keys[i] < element ) i++;
  return i;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::Display --
 *
 *  Show all the data relative to the node on the channel os
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
void 
BTreeNode<TYPE>::Display(ostream& os) {
  os << "[" << (m_links == NULL ? "O" : "X") 
     << "," << m_currPos
     << "," << m_nElement
     << "," << m_arity
     << "] = " << *this << "\n";
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::operator<< --
 *
 *  Show contents of the node <node> on the stream <os>
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
ostream& 
operator<<(ostream& os, const BTreeNode<TYPE>& node) {

  os << (node.IsLeaf() ? "(" : "[");
  for (int i = 0; i < node.m_nElement; i++) {
    os << node.m_keys[i] << " ";
  }
  os << (node.m_nElement != 0 ? "\b" : "") 
     <<	 (node.IsLeaf() ? ")" : "]");
  return os;
}

/*
 *------------------------------------------------------------------------
 *
 *  BTreeNode::Append --
 *
 *  Append a new element to the node;
 *  Parameter:
 *   element (input by reference) the element to append
 *  Result:
 *   None
 *  Side effect:
 *   Increment the number of element present
 *
 *------------------------------------------------------------------------
 */
template<class TYPE>  
void
BTreeNode<TYPE>::Append(const TYPE& element) {
  if (m_nElement == m_arity) {
    cerr << "Cant't append element!\n";
    return;
  }
  m_keys[m_nElement] = element;
  m_nElement++;
}

/*
 *--------------------------------------------------------------
 *
 * BTreeNo::Save --
 *
 *	Save the tree into a stream
 *
 * Arguments:
 *	ofstream &os: the stream referred
 * Results:
 *	Returns void.
 *
 *--------------------------------------------------------------
 */
template<class TYPE>  
void
BTreeNode<TYPE>::Save (ostream &os) {

  for (int i = 0; i < m_nElement; i++) {
    os << m_keys[i] << "*\n";
  }
  return os;
  
}

#endif /* BTREENODE_DEFINITIONS_INCLUDED */
